# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from munch import Munch

BODHI_LATEST_BUILDS = Munch(
    {
        "f28-updates-candidate": "colin-0.3.1-1.fc28",
        "f29-updates-candidate": "colin-0.3.1-1.fc29",
        "f30-updates-candidate": "colin-0.3.1-2.fc30",
        "f27-updates-candidate": "colin-0.2.0-1.fc27",
        "f28-updates-testing": "colin-0.3.1-1.fc28",
        "f29-updates-testing": "colin-0.3.1-1.fc29",
        "f27-updates-testing": "colin-0.2.0-1.fc27",
        "f30-updates-testing": "colin-0.3.1-2.fc30",
        "f28-updates": "colin-0.3.1-1.fc28",
        "f29-updates": "colin-0.3.1-1.fc29",
        "f27-updates": "colin-0.2.0-1.fc27",
        "f30-updates": "colin-0.3.1-2.fc30",
        "f28-override": "colin-0.3.1-1.fc28",
        "f29-override": "colin-0.3.1-1.fc29",
        "f27-override": "colin-0.2.0-1.fc27",
        "f30-override": "colin-0.3.1-2.fc30",
        "f28-updates-testing-pending": "colin-0.3.1-1.fc28",
        "f29-updates-testing-pending": "colin-0.3.1-1.fc29",
        "f27-updates-testing-pending": "colin-0.0.4-3.fc27",
        "f30-updates-testing-pending": "colin-0.3.1-2.fc30",
        "f28-updates-pending": "colin-0.3.1-1.fc28",
        "f29-updates-pending": "colin-0.3.1-1.fc29",
        "f27-updates-pending": "colin-0.2.0-1.fc27",
        "f30-updates-pending": "colin-0.3.1-2.fc30",
    },
)
